import torch
from torch import nn
from zeta.nn.attention import Attention
from zeta.nn import FeedForward
from zeta.nn import audio_to_text, img_to_text, video_to_text


class RekaTransformerBlock(nn.Module):
    """
    RekaTransformerBlock is a module that represents a single transformer block in the RekaTransformer model.

    Args:
        dim (int): The input dimension of the transformer block.
        depth (int): The number of layers in the transformer block.
        dim_head (int, optional): The dimension of each head in the multi-head attention mechanism. Defaults to 64.
        heads (int, optional): The number of attention heads. Defaults to 8.
        ff_mult (int, optional): The multiplier for the hidden dimension in the feed-forward network. Defaults to 4.
        ff_dropout (float, optional): The dropout probability for the feed-forward network. Defaults to 0.0.
    """

    def __init__(
        self,
        dim: int,
        dim_head: int = 64,
        heads: int = 8,
        ff_mult: int = 4,
        ff_dropout: float = 0.0,
    ):
        super().__init__()
        self.dim = dim
        self.dim_head = dim_head
        self.heads = heads
        self.ff_mult = ff_mult
        self.ff_dropout = ff_dropout

        # Attention Layers
        self.attn = Attention(
            dim=dim,
            dim_head=dim_head,
            heads=heads,
            causal=True,
            qk_norm=True,
            kv_heads=4,
        )

        # Feed Forward Layers
        self.ffn = FeedForward(
            dim,
            dim,
            mult=ff_mult,
            post_act_ln=True,
            swish=True,
        )

    def forward(self, x):
        """
        Forward pass of the RekaTransformerBlock.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The output tensor.
        """
        skip = x
        x, _ = self.attn(x)
        # print(x.shape, skip.shape)
        attended = x + skip

        x = self.ffn(attended)
        x += skip
        return x


class RekaTransformer(nn.Module):
    def __init__(
        self,
        dim: int,
        depth: int = 6,
        dim_head: int = 64,
        heads: int = 8,
        ff_mult: int = 4,
        ff_dropout: float = 0.0,
    ):
        super().__init__()
        self.dim = dim
        self.dim_head = dim_head
        self.heads = heads
        self.ff_mult = ff_mult
        self.ff_dropout = ff_dropout

        self.blocks = nn.ModuleList(
            [
                RekaTransformerBlock(
                    dim=dim,
                    dim_head=dim_head,
                    heads=heads,
                    ff_mult=ff_mult,
                    ff_dropout=ff_dropout,
                )
                for _ in range(depth)
            ]
        )

    def forward(self, x):
        for block in self.blocks:
            x = block(x)
        return x


class Reka(nn.Module):
    """
    Reka model for multimodal fusion.

    Args:
        dim (int): Dimension of the input features.
        depth (int, optional): Number of transformer layers. Defaults to 6.
        dim_head (int, optional): Dimension of each head in the transformer. Defaults to 64.
        heads (int, optional): Number of attention heads in the transformer. Defaults to 8.
        ff_mult (int, optional): Multiplier for the feed-forward layer dimension. Defaults to 4.
        ff_dropout (float, optional): Dropout probability for the feed-forward layer. Defaults to 0.0.
    """

    def __init__(
        self,
        dim: int,
        max_seq_len: int = 512,
        depth: int = 6,
        dim_head: int = 64,
        heads: int = 8,
        ff_mult: int = 4,
        ff_dropout: float = 0.0,
        post_modal_transform_norm: bool = True,
        post_fusion_norm: bool = True,
        vocab_size: int = 10000,
    ):
        super().__init__()
        self.dim = dim
        self.max_seq_len = max_seq_len
        self.dim_head = dim_head
        self.heads = heads
        self.ff_mult = ff_mult
        self.ff_dropout = ff_dropout
        self.post_modal_transform_norm = post_modal_transform_norm
        self.post_fusion_norm = post_fusion_norm

        self.transformer = RekaTransformer(
            dim=dim,
            depth=depth,
            dim_head=dim_head,
            heads=heads,
            ff_mult=ff_mult,
            ff_dropout=ff_dropout,
        )

        # Post fusion norm
        if self.post_fusion_norm:
            self.psf_norm = nn.LayerNorm(dim)

        if self.post_modal_transform_norm:
            self.pmt_norm = nn.LayerNorm(dim)

        # Embedder
        self.token = nn.Embedding(vocab_size, dim)

    def forward(self, text, img, audio, video):
        """
        Forward pass of the Reka model.

        Args:
            text: Not used in this implementation.
            img (torch.Tensor): Input image tensor.
            audio (torch.Tensor): Input audio tensor.
            video (torch.Tensor): Input video tensor.

        Returns:
            torch.Tensor: Output tensor after the forward pass.
        """
        logits = self.token(text)

        assert (
            (img is not None and audio is not None)
            or (img is not None and video is not None)
            or (audio is not None and video is not None)
        ), (
            "At least two of the inputs (img, audio, video) must be"
            " provided."
        )

        if img is not None:
            # Image dimensions
            img_b, img_c, img_h, img_w = img.shape

            # img = img_to_text(img, self.patches, self.patch_size, self.dim, True)
            img = img_to_text(img, self.max_seq_len, self.dim, True)

            if self.post_modal_transform_norm:
                img = self.pmt_norm(img)

        if audio is not None:
            # Audio dimensions
            audio_b, audio_seq_len = audio.shape

            audio = audio_to_text(
                audio, self.max_seq_len, self.dim, True
            )

            if self.post_modal_transform_norm:
                audio = self.pmt_norm(audio)

        if video is not None:
            # Video dimensions
            video_b, video_c, video_f, video_h, video_w = video.shape

            video = video_to_text(
                video, self.max_seq_len, self.dim, True
            )

        # Fuse layers
        if img is not None and audio is not None:
            fused = torch.cat((img, audio), dim=1)
        elif img is not None and video is not None:
            fused = torch.cat((img, video), dim=1)
        elif audio is not None and video is not None:
            fused = torch.cat((audio, video), dim=1)

        # Post fusion layernorm for stability.
        if self.post_fusion_norm:
            fused = self.psf_norm(fused)

        # Fuse
        fused = torch.cat((logits, fused), dim=1)

        # Transformer
        x = self.transformer(fused)

        return x



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['reka_torch']

package_data = \
{'': ['*']}

install_requires = \
['einops', 'torch', 'torchvision', 'zetascale']

setup_kwargs = {
    'name': 'reka-torch',
    'version': '0.0.2',
    'description': 'Reka Torch - Pytorch',
    'long_description': '[![Multi-Modality](agorabanner.png)](https://discord.gg/qUtxnK2NMf)\n\n# Reka Torch\nImplementation of the model: "Reka Core, Flash, and Edge: A Series of Powerful Multimodal Language Models" in PyTorch. [PAPER LINK](https://publications.reka.ai/reka-core-tech-report.pdf)\n\n## Install\n`pip3 install -U reka-torch`\n\n## Usage\n```python\nimport torch  # Importing the torch library\nfrom reka_torch.model import Reka  # Importing the Reka model from the reka_torch package\n\ntext = torch.randint(0, 10000, (2, 512))  # Generating a random tensor of shape (2, 512) with values between 0 and 10000\n\nimg = torch.randn(2, 3, 224, 224)  # Generating a random tensor of shape (2, 3, 224, 224) with values from a normal distribution\n\naudio = torch.randn(2, 1000)  # Generating a random tensor of shape (2, 1000) with values from a normal distribution\n\nvideo = torch.randn(2, 3, 16, 224, 224)  # Generating a random tensor of shape (2, 3, 16, 224, 224) with values from a normal distribution\n\nmodel = Reka(512)  # Creating an instance of the Reka model with input size 512\n\nout = model(text, img, audio, video)  # Forward pass through the model with the input tensors\n\nprint(out.shape)  # Printing the shape of the output tensor\n\n```\n\n# License\nMIT\n',
    'author': 'Kye Gomez',
    'author_email': 'kye@apac.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/kyegomez/Reka-Torch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
